/*
 * Copyright (c) 1997, 1998  Motoyuki Kasahara
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <sys/types.h>
#include <ctype.h>

#if defined(STDC_HEADERS) || defined(HAVE_STRING_H)
#include <string.h>
#if !defined(STDC_HEADERS) && defined(HAVE_MEMORY_H)
#include <memory.h>
#endif /* not STDC_HEADERS and HAVE_MEMORY_H */
#else /* not STDC_HEADERS and not HAVE_STRING_H */
#include <strings.h>
#endif /* not STDC_HEADERS and not HAVE_STRING_H */

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#else
#include <sys/file.h>
#endif

#ifdef HAVE_LIMITS_H
#include <limits.h>
#endif

#include "eb.h"
#include "error.h"
#include "io.h"
#include "appendix.h"

#ifndef HAVE_STRCHR
#define strchr index
#define strrchr rindex
#endif /* HAVE_STRCHR */

#ifndef O_BINARY
#define O_BINARY 0
#endif

/*
 * Whence parameter for lseek().
 */
#ifndef SEEK_SET
#define SEEK_SET 0
#define SEEK_CUR 1
#define SEEK_END 2
#endif

/*
 * The maximum length of path name.
 */
#ifndef PATH_MAX
#ifdef MAXPATHLEN
#define PATH_MAX	MAXPATHLEN
#else /* not MAXPATHLEN */
#define PATH_MAX	1024
#endif /* not MAXPATHLEN */
#endif /* not PATH_MAX */

#ifdef __STDC__
static void eb_initialize_alternation_cache(EB_Appendix *);
static int eb_initialize_appendix_catalog(EB_Appendix *);
#else
static int eb_initialize_appendix_catalog();
static void eb_initialize_alternation_cache();
#endif


/*
 * Clear cache for alternation text in `appendix'.
 */
static void
eb_initialize_alternation_cache(appendix)
    EB_Appendix *appendix;
{
    EB_Alternation_Cache *p;
    int i;

    for (i = 0, p = appendix->narwcache;
	 i < EB_MAX_ALTERNATION_CACHE; i++, p++)
	p->charno = -1;
    for (i = 0, p = appendix->widecache;
	 i < EB_MAX_ALTERNATION_CACHE; i++, p++)
	p->charno = -1;
}


/*
 * Initialize `appendix'.
 */
void
eb_initialize_appendix(appendix)
    EB_Appendix *appendix;
{
    appendix->path = NULL;
    appendix->subcurrent = NULL;
    appendix->subbooks = NULL;

    eb_initialize_alternation_cache(appendix);
}


/*
 * Finish using `appendix'.
 */
void
eb_clear_appendix(appendix)
    EB_Appendix *appendix;
{
    if (appendix->subbooks != NULL)
	free(appendix->subbooks);

    if (appendix->path != NULL)
	free(appendix->path);

    eb_initialize_appendix(appendix);
}


/*
 * Suspend using `appendix'.
 */
void
eb_suspend_appendix(appendix)
    EB_Appendix *appendix;
{
    eb_unset_appendix_subbook(appendix);
}


/*
 * Bind `appendix' to `path'.
 */
int
eb_bind_appendix(appendix, path)
    EB_Appendix *appendix;
    const char *path;
{
    char tmppath[PATH_MAX + 1];
    int count;

    /*
     * Reset structure members in the appendix.
     */
    eb_initialize_appendix(appendix);

    /*
     * Set a path of the appendix.
     * The length of the filename "path/subdir/subsubdir/file.;1" must
     * not exceed PATH_MAX.
     */
    if (PATH_MAX < strlen(path)) {
	eb_error = EB_ERR_TOO_LONG_FNAME;
	goto failed;
    }

    strcpy(tmppath, path);
    if (eb_canonicalize_filename(tmppath) < 0)
	goto failed;

    appendix->pathlen = strlen(tmppath);
    if (PATH_MAX < appendix->pathlen + (1 + EB_MAXLEN_BASENAME + 1
	+ EB_MAXLEN_BASENAME + 1 + EB_MAXLEN_BASENAME + 3)) {
	eb_error = EB_ERR_TOO_LONG_FNAME;
	goto failed;
    }

    appendix->path = (char *)malloc(appendix->pathlen + 1);
    if (appendix->path == NULL) {
	eb_error = EB_ERR_MEMORY_EXHAUSTED;
	goto failed;
    }
    strcpy(appendix->path, tmppath);

    /*
     * Get disc type and filename mode.
     */
    if (eb_appendix_catalog_filename(appendix) < 0)
	goto failed;

    /*
     * Read information from the `CATALOG(S)' file.
     */
    count = eb_initialize_appendix_catalog(appendix);
    if (count < 0)
	goto failed;
    
    return count;

    /*
     * An error occurs...
     */
  failed:
    eb_clear_appendix(appendix);
    return -1;
}


/*
 * Read information from the `CATALOG(S)' file in `appendix'.
 *
 * If succeeded, the number of subbooks in `appendix' is returned.
 * Otherwise, -1 is returned and `eb_error' is set.
 */
static int
eb_initialize_appendix_catalog(appendix)
    EB_Appendix *appendix;
{
    char buf[EB_SIZE_PAGE];
    char catalog[PATH_MAX + 1];
    char *space;
    EB_Appendix_Subbook *sub;
    size_t catalogsize;
    size_t titlesize;
    int file = -1;
    int i;

    if (appendix->disccode == EB_DISC_EB) {
	catalogsize = EB_SIZE_EB_CATALOG;
	titlesize = EB_MAXLEN_EB_TITLE;
	sprintf(catalog, "%s/%s", appendix->path, EB_FNAME_CATALOG);
    } else {
	catalogsize = EB_SIZE_EP_CATALOG;
	titlesize = EB_MAXLEN_EP_TITLE;
	sprintf(catalog, "%s/%s", appendix->path, EB_FNAME_CATALOGS);
    }
	
    /*
     * Open the catalog file.
     */
    eb_fix_appendix_filename(appendix, catalog);
    file = open(catalog, O_RDONLY | O_BINARY);
    if (file < 0)
	goto failed;

    /*
     * Get the number of subbooks in the appendix.
     */
    if (read(file, buf, 16) != 16) {
	eb_error = EB_ERR_FAIL_READ_CATAPP;
	goto failed;
    }
    appendix->subcount = eb_uint2(buf);
    if (EB_MAX_SUBBOOKS < appendix->subcount)
	appendix->subcount = EB_MAX_SUBBOOKS;

    /*
     * Allocate memories for subbook entries.
     */
    appendix->subbooks = (EB_Appendix_Subbook *)
	malloc(sizeof(EB_Appendix_Subbook) * appendix->subcount);
    if (appendix->subbooks == NULL) {
	eb_error = EB_ERR_MEMORY_EXHAUSTED;
	goto failed;
    }

    /*
     * Read subbook information.
     */
    for (i = 0, sub = appendix->subbooks; i < appendix->subcount; i++, sub++) {
	/*
	 * Read data from the catalog file.
	 */
	if (read(file, buf, catalogsize) != catalogsize) {
	    eb_error = EB_ERR_FAIL_READ_CAT;
	    goto failed;
	}

	/*
	 * Set a directory name of the subbook.
	 */
	strncpy(sub->directory, buf + 2 + titlesize, EB_MAXLEN_BASENAME);
	sub->directory[EB_MAXLEN_BASENAME] = '\0';
	space = strchr(sub->directory, ' ');
	if (space != NULL)
	    *space = '\0';

	sub->initialized = 0;
	sub->code = i;
    }

    /*
     * Close the catalog file.
     */
    close(file);
    return appendix->subcount;

    /*
     * An error occurs...
     */
  failed:
    if (0 <= file)
	close(file);

    if (appendix->subbooks != NULL) {
	free(appendix->subbooks);
	appendix->subbooks = NULL;
    }

    return -1;
}


/*
 * Read information from the subbooks in `appendix'.
 */
int
eb_initialize_appendix_subbook(appendix)
    EB_Appendix *appendix;
{
    char buf[16];
    int stoppage;
    int len;

    /*
     * Current subbook must have been set.
     */
    if (appendix->subcurrent == NULL) {
	eb_error = EB_ERR_NO_CUR_APPSUB;
	return -1;
    }

    /*
     * If the subbook has already initialized, return immediately.
     */
    if (appendix->subcurrent->initialized != 0)
	return 0;

    /*
     * Rewind the APPENDIX file.
     */
    if (lseek(appendix->subcurrent->appfile, 0, SEEK_SET) < 0) {
	eb_error = EB_ERR_FAIL_SEEK_APP;
	return -1;
    }

    /*
     * Set general information about appendix the subbook.
     */
    if (read(appendix->subcurrent->appfile, buf, 16) != 16) {
	eb_error = EB_ERR_FAIL_READ_APP;
	return -1;
    }
    appendix->subcurrent->charcode = eb_uint2(buf + 2);

    /*
     * Set information about alternation text for a narrow font.
     */
    if (read(appendix->subcurrent->appfile, buf, 16) != 16) {
	eb_error = EB_ERR_FAIL_READ_APP;
	return -1;
    }
    len = eb_uint2(buf + 12);
    appendix->subcurrent->narwpage = eb_uint4(buf);
    appendix->subcurrent->narwstart = eb_uint2(buf + 10);
    appendix->subcurrent->narwend = appendix->subcurrent->narwstart
	+ ((len / 0x5e) << 8) + (len % 0x5e) - 1;
    if (0x7e < (appendix->subcurrent->narwend & 0xff))
	appendix->subcurrent->narwend += 0xa3;

    /*
     * Set information about alternation text for a wide font.
     */
    if (read(appendix->subcurrent->appfile, buf, 16) != 16) {
	eb_error = EB_ERR_FAIL_READ_APP;
	return -1;
    }
    len = eb_uint2(buf + 12);
    appendix->subcurrent->widepage = eb_uint4(buf);
    appendix->subcurrent->widestart = eb_uint2(buf + 10);
    appendix->subcurrent->wideend = appendix->subcurrent->widestart
	+ ((len / 0x5e) << 8) + (len % 0x5e) - 1;
    if (0x7e < (appendix->subcurrent->wideend & 0xff))
	appendix->subcurrent->wideend += 0xa3;
    
    /*
     * Set a stop-code;
     */
    if (read(appendix->subcurrent->appfile, buf, 16) != 16) {
	eb_error = EB_ERR_FAIL_READ_APP;
	return -1;
    }
    stoppage = eb_uint4(buf);
    if (lseek(appendix->subcurrent->appfile, (stoppage - 1) * EB_SIZE_PAGE,
	SEEK_SET) < 0) {
	eb_error = EB_ERR_FAIL_SEEK_APP;
	return -1;
    }
    if (read(appendix->subcurrent->appfile, buf, 16) != 16) {
	eb_error = EB_ERR_FAIL_READ_APP;
	return -1;
    }
    if (eb_uint2(buf) != 0) {
	appendix->subcurrent->stop0 = eb_uint2(buf + 2);
	appendix->subcurrent->stop1 = eb_uint2(buf + 4);
    } else {
	appendix->subcurrent->stop0 = 0;
	appendix->subcurrent->stop1 = 0;
    }

    /*
     * Rewind the file descriptor of the APPENDIX file.
     */
    if (lseek(appendix->subcurrent->appfile, 0, SEEK_SET) < 0) {
	eb_error = EB_ERR_FAIL_SEEK_APP;
	return -1;
    }

    eb_initialize_alternation_cache(appendix);

    return 0;
}


/*
 * Get information about all subbooks in the book.
 */
int
eb_initialize_all_appendix_subbooks(appendix)
    EB_Appendix *appendix;
{
    EB_Subbook_Code cur;
    EB_Appendix_Subbook *sub;
    int i;

    /*
     * Get the current subbook.
     */
    if (appendix->subcurrent != NULL)
	cur = appendix->subcurrent->code;
    else
	cur = -1;

    /*
     * Initialize each subbook.
     */
    for (i = 0, sub = appendix->subbooks; i < appendix->subcount; i++, sub++) {
	if (eb_set_appendix_subbook(appendix, sub->code) < 0)
	    return -1;
    }

    /*
     * Restore the current subbook.
     */
    if (cur < 0)
	eb_unset_appendix_subbook(appendix);
    else if (eb_set_appendix_subbook(appendix, cur) < 0)
	return -1;

    return 0;
}


/*
 * Examine whether `appendix' is bound to a path or not.
 */
int
eb_is_appendix_bound(appendix)
    EB_Appendix *appendix;
{
    if (appendix->path == NULL) {
	eb_error = EB_ERR_UNBOUND_APP;
	return 0;
    }

    return 1;
}


/*
 * Return the bound path of `appendix'.
 */
const char *
eb_appendix_path(appendix)
    EB_Appendix *appendix;
{
    /*
     * The appendix must have been bound.
     */
    if (appendix->path == NULL) {
	eb_error = EB_ERR_UNBOUND_APP;
	return NULL;
    }

    return appendix->path;
}


/*
 * Return the number of subbooks in `appendix'.
 */
int
eb_appendix_subbook_count(appendix)
    EB_Appendix *appendix;
{
    /*
     * The book must have been bound.
     */
    if (appendix->path == NULL) {
	eb_error = EB_ERR_UNBOUND_APP;
	return -1;
    }

    return appendix->subcount;
}


/*
 * Make a subbook list in `appendix'.
 */
int
eb_appendix_subbook_list(appendix, list)
    EB_Appendix *appendix;
    EB_Subbook_Code *list;
{
    EB_Subbook_Code *lp;
    int i;

    /*
     * The book must have been bound.
     */
    if (appendix->path == NULL) {
	eb_error = EB_ERR_UNBOUND_APP;
	return -1;
    }

    for (i = 0, lp = list; i < appendix->subcount; i++, lp++)
	*lp = i;

    return appendix->subcount;
}


/*
 * Return the subbook-code of the current subbook in `appendix'.
 */
EB_Subbook_Code
eb_appendix_subbook(appendix)
    EB_Appendix *appendix;
{
    /*
     * Current subbook must have been set.
     */
    if (appendix->subcurrent == NULL) {
	eb_error = EB_ERR_NO_CUR_APPSUB;
	return -1;
    }

    return appendix->subcurrent->code;
}


/*
 * Return the directory name of the current subbook in `appendix'.
 */
const char *
eb_appendix_subbook_directory(appendix)
    EB_Appendix *appendix;
{
    /*
     * Current subbook must have been set.
     */
    if (appendix->subcurrent == NULL) {
	eb_error = EB_ERR_NO_CUR_APPSUB;
	return NULL;
    }

    return appendix->subcurrent->directory;
}


/*
 * Return the directory name of the subbook `code' in `appendix'.
 */
const char *
eb_appendix_subbook_directory2(appendix, code)
    EB_Appendix *appendix;
    EB_Subbook_Code code;
{
    /*
     * The book must have been bound.
     */
    if (appendix->path == NULL) {
	eb_error = EB_ERR_UNBOUND_APP;
	return NULL;
    }

    /*
     * Check for the subbook-code.
     */
    if (code < 0 || appendix->subcount <= code) {
	eb_error = EB_ERR_NO_SUCH_APPSUB;
	return NULL;
    }

    return (appendix->subbooks + code)->directory;
}


/*
 * Set the subbook `code' as the current subbook.
 */
int
eb_set_appendix_subbook(appendix, code)
    EB_Appendix *appendix;
    EB_Subbook_Code code;
{
    char filename[PATH_MAX + 1];

    /*
     * The book must have been bound.
     */
    if (appendix->path == NULL) {
	eb_error = EB_ERR_UNBOUND_APP;
	goto failed;
    }

    /*
     * Check for the subbook-code.
     */
    if (code < 0 || appendix->subcount <= code) {
	eb_error = EB_ERR_NO_SUCH_APPSUB;
	goto failed;
    }

    /*
     * If the current subbook is the subbook with `code', return
     * immediately.
     * Otherwise close the current subbook and continue.
     */
    if (appendix->subcurrent != NULL) {
	if (appendix->subcurrent->code == code)
	    return 0;
	eb_unset_appendix_subbook(appendix);
    }

    /*
     * Open the subbook.
     */
    appendix->subcurrent = appendix->subbooks + code;
    if (appendix->disccode == EB_DISC_EB) {
	sprintf(filename, "%s/%s/%s", appendix->path,
	    appendix->subcurrent->directory, EB_FNAME_APPENDIX);
    } else {
	sprintf(filename, "%s/%s/%s/%s", appendix->path,
	    appendix->subcurrent->directory, EB_DIRNAME_DATA, EB_FNAME_FUROKU);
    }
    eb_fix_appendix_filename(appendix, filename);
    appendix->subcurrent->appfile = open(filename, O_RDONLY | O_BINARY);
    if (appendix->subcurrent->appfile < 0) {
	eb_error = EB_ERR_FAIL_OPEN_APP;
	appendix->subcurrent = NULL;
	goto failed;
    }

    /*
     * Initialize the subbook.
     */
    if (eb_initialize_appendix_subbook(appendix) < 0)
	goto failed;

    return 0;

    /*
     * An error occurs...
     */
  failed:
    eb_unset_appendix_subbook(appendix);
    return -1;
}


/*
 * Unset the current subbook.
 */
void
eb_unset_appendix_subbook(appendix)
    EB_Appendix *appendix;
{
    /*
     * Close the file of the current subbook.
     */
    if (appendix->subcurrent != NULL) {
	close(appendix->subcurrent->appfile);
	appendix->subcurrent = NULL;
    }
}


