/*
 * Copyright (c) 1997  Motoyuki Kasahara
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>

#if defined(STDC_HEADERS) || defined(HAVE_STRING_H)
#include <string.h>
#if !defined(STDC_HEADERS) && defined(HAVE_MEMORY_H)
#include <memory.h>
#endif /* not STDC_HEADERS and HAVE_MEMORY_H */
#else /* not STDC_HEADERS and not HAVE_STRING_H */
#include <strings.h>
#endif /* not STDC_HEADERS and not HAVE_STRING_H */

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifndef HAVE_STRCASECMP
#ifdef __STDC__
int strcasecmp(const char *, const char *);
int strncasecmp(const char *, const char *, size_t);
#else /* not __STDC__ */
int strcasecmp();
int strncasecmp();
#endif /* not __STDC__ */
#endif /* not HAVE_STRCASECMP */

#include <eb/eb.h>
#include <eb/appendix.h>
#include <eb/language.h>

#include "wildcard.h"
#include "ndtpd.h"
#include "permission.h"


/*
 * Initialize the book list `book_registry'.
 */
void
initialize_book_registry()
{
    book_registry = NULL;
    book_count = 0;
    current_book = NULL;
}


/*
 * Clear the book list `book_registry'.
 * Free all nodes and allocated memories in `book_registry',
 * and then set it to NULL.
 */
void
clear_book_registry()
{
    Book *book = book_registry;
    Book *next;

    /*
     * Reset books list.  Free all malloced memories.
     */
    while (book != NULL) {
	clear_permission(&book->permissions);
	eb_clear(&book->book);
	eb_clear_appendix(&book->appendix);
	next = (Book *)book->next;
	free(book);
	book = next;
    }

    book_registry = NULL;
    current_book = NULL;
    book_count = 0;
}


/*
 * Terminate the book list `book_registry'.
 */
void
terminate_book_registry()
{
    Book *book;

    /*
     * Initialize all subbooks when running as a standalone daemon.
     */
    if (server_mode == SERVER_MODE_STANDALONE) {
	for (book = book_registry; book != NULL; book = (Book *)book->next) {
	    eb_initialize_all_subbooks(&book->book);
	    eb_initialize_all_appendix_subbooks(&book->appendix);
	}
    }

    /*
     * Unset the current book.
     */
    current_book = NULL;
}


/*
 * Add a new book to the book list `book_registry'.
 * Upon return, `current_book' refers to the new book.
 * If succeeded, the added node is returned.
 * Otherwise NULL is returned.
 */
int
add_book()
{
    Book *book;

    book = (Book *)malloc(sizeof(Book));
    if (book == NULL)
	return -1;

    /*
     * Set initial values to the new book.
     */
    eb_initialize(&book->book);
    eb_initialize_appendix(&book->appendix);
    book->code = -1;
    book->name[0] = '\0';
    book->title[0] = '\0';
    book->max_clients = 0;
    initialize_permission(&book->permissions);
    book->next = book_registry;
    book_registry = book;
    current_book = book;

    book_count++;

    return 0;
}


/*
 * Delete the current book, added by add_book().
 */
void
delete_book()
{
    if (current_book == NULL)
	return;

    clear_permission(&current_book->permissions);
    eb_clear(&current_book->book);
    eb_clear_appendix(&current_book->appendix);
    book_registry = (Book *)current_book->next;
    free(current_book);
    current_book = NULL;
}


/*
 * Set the current book.
 *
 * The pointer to the book with `name' is returned.
 * If no such book in `book_registry', NULL is returned.
 */
int
set_current_book(name)
    const char *name;
{
    /*
     * If the current book is active, unset it.
     */
    if (current_book != NULL) {
	if (strcasecmp(current_book->name, name) == 0)
	    return 0;

	eb_suspend(&current_book->book);
	eb_suspend_appendix(&current_book->appendix);
	current_book = NULL;
    }

    /*
     * Find the book.
     */
    current_book = find_book(name);
    if (current_book == NULL)
	return -1;

    current_book->subname = NULL;
    return 0;
}


/*
 * Unset the current subbook.
 */
void
unset_current_book()
{
    if (current_book == NULL)
	return;

    eb_suspend(&current_book->book);
    eb_suspend_appendix(&current_book->appendix);
    current_book = NULL;
}


/*
 * Set the current subbook.
 *
 * If succeeded, the pointer to the current book is returned.
 * Otherwise, NULL is returned.
 */
int
set_current_subbook(name)
    const char *name;
{
    EB_Subbook_Code sublist[EB_MAX_SUBBOOKS];
    const char *dir;
    int count;
    int i;

    /*
     * The current book must have been selected.
     */
    if (current_book == NULL)
	return -1;

    /*
     * If the current subbook is active, unset it.
     */
    if (current_book->subname != NULL) {
	if (strcasecmp(current_book->subname, name) == 0)
	    return 0;
	eb_unset_subbook(&current_book->book);
	eb_unset_appendix_subbook(&current_book->appendix);
	current_book->subname = NULL;
    }

    /*
     * Find the subbook in the current book.
     */
    count = eb_subbook_list(&current_book->book, sublist);
    for (i = 0; i < count; i++) {
	dir = eb_subbook_directory2(&current_book->book, sublist[i]);
	if (strcasecmp(name, dir) == 0) {
	    if (eb_set_subbook(&current_book->book, sublist[i]) < 0) {
		unset_current_book();
		return -1;
	    }
	    current_book->subname = dir;
	    break;
	}
    }
    if (count <= i) {
	eb_error = EB_ERR_NO_SUCH_SUB;
	unset_current_book();
	return -1;
    }

    /*
     * Find the subbook in the current appendix.
     */
    if (eb_is_appendix_bound(&current_book->appendix)) {
	count = eb_appendix_subbook_list(&current_book->appendix, sublist);
	for (i = 0; i < count; i++) {
	    dir = eb_appendix_subbook_directory2(&current_book->appendix,
		sublist[i]);
	    if (strcasecmp(name, dir) == 0) {
		if (eb_set_appendix_subbook(&current_book->appendix,
		    sublist[i]) < 0) {
		    unset_current_book();
		    return -1;
		}
		break;
	    }
	}
	if (count <= i) {
	    eb_error = EB_ERR_NO_SUCH_APPSUB;
	    unset_current_book();
	    return -1;
	}
    }

    return 0;
}


/*
 * Unset the current subbook.
 */
void
unset_current_subbook()
{
    if (current_book == NULL)
	return;

    eb_unset_subbook(&current_book->book);
    eb_unset_appendix_subbook(&current_book->appendix);
    current_book->subname = NULL;
}


/*
 * Find the book with `name' in `book_regitry'.
 *
 * If found, the pointer to the book is returned.
 * Otherwise NULL is returned.
 */
Book *
find_book(name)
    const char *name;
{
    Book *book;

    /*
     * Find the book.
     */
    for (book = book_registry; book != NULL; book = (Book *)book->next) {
	if (strcasecmp(book->name, name) == 0)
	    return book;
    }

    return NULL;
}


/*
 * Set permission flags to the current client.
 */
void
set_book_permissions()
{
    Book *book;

    for (book = book_registry; book != NULL; book = (Book *)book->next) {
	book->permflag = test_permission(&book->permissions, client_hostname,
	    client_address, match_wildcard);
    }
}
